# -*- encoding:utf-8 -*-
# A dispatcher is a class that can be used to send and receive messages between pqi-server and single pqi-client.

from PyQt5 import QtCore
import threading

from PyQtInspect._pqi_bundle.pqi_comm import ReaderThread, WriterThread, NetCommandFactory
from PyQtInspect._pqi_bundle.pqi_override import overrides


class DispatchReader(ReaderThread):
    def __init__(self, dispatcher):
        ReaderThread.__init__(self, dispatcher.sock)
        self.dispatcher = dispatcher

    @overrides(ReaderThread._on_run)
    def _on_run(self):
        dummy_thread = threading.current_thread()
        dummy_thread.is_pydev_daemon_thread = False
        return ReaderThread._on_run(self)

    def handle_except(self):
        self.dispatcher.notifyDelete()

    def process_command(self, cmd_id, seq, text):
        # unquote text
        from urllib.parse import unquote
        text = unquote(text)
        self.dispatcher.notify(cmd_id, seq, text)


class Dispatcher(QtCore.QThread):
    sigMsg = QtCore.pyqtSignal(int, dict)  # dispatcher_id, info
    sigDelete = QtCore.pyqtSignal(int)

    def __init__(self, parent, sock, id):
        super().__init__(parent)
        self.sock = sock
        self.id = id
        self.net_command_factory = NetCommandFactory()
        self.reader = None
        self.writer = None

        # 由于dispatcher刚创建的时候, 主界面还来不及处理立即到来的信息(PQYWorker未发出新dispatcher创建的信号)
        # 所以需要在主界面准备好之后再处理
        # 此前的消息都缓存起来
        self._mainUIReady = False
        self._msg_buffer = []

    def run(self):
        self.writer = WriterThread(self.sock)
        self.writer.pydev_do_not_trace = False  # We run writer in the same thread so we don't want to loose tracing.
        self.writer.start()

        self.reader = DispatchReader(self)
        self.reader.pydev_do_not_trace = False  # We run reader in the same thread so we don't want to loose tracing.
        self.reader.run()

    def close(self):
        try:
            self.writer.do_kill_pydev_thread()
            self.reader.do_kill_pydev_thread()
            self.sock.close()
        except:
            pass

    def registerMainUIReady(self):
        """ 告知dispatcher, 主界面已经准备好了, 可以处理消息了
        """
        self._mainUIReady = True
        for cmd_id, seq, text in self._msg_buffer:
            self.notify(cmd_id, seq, text)

    def notify(self, cmd_id, seq, text):
        if not self._mainUIReady:
            # 缓存起来
            self._msg_buffer.append((cmd_id, seq, text))
        self.sigMsg.emit(self.id, {"cmd_id": cmd_id, "seq": seq, "text": text})

    def sendEnableInspect(self, extra: dict):
        self.writer.add_command(self.net_command_factory.make_enable_inspect_message(extra))

    def sendDisableInspect(self):
        self.writer.add_command(self.net_command_factory.make_disable_inspect_message())

    def sendExecCodeEvent(self, code: str):
        self.writer.add_command(self.net_command_factory.make_exec_code_message(code))

    def sendHighlightWidgetEvent(self, widgetId: int, isHighlight: bool):
        self.writer.add_command(self.net_command_factory.make_set_widget_highlight_message(widgetId, isHighlight))

    def sendSelectWidgetEvent(self, widgetId: int):
        self.writer.add_command(self.net_command_factory.make_select_widget_message(widgetId))

    def sendRequestWidgetInfoEvent(self, widgetId: int, extra: dict = None):
        self.writer.add_command(self.net_command_factory.make_req_widget_info_message(widgetId, extra))

    def sendRequestChildrenInfoEvent(self, widgetId: int):
        self.writer.add_command(self.net_command_factory.make_req_children_info_message(widgetId))

    def notifyDelete(self):
        self.reader.do_kill_pydev_thread()
        self.writer.do_kill_pydev_thread()

        self.sigDelete.emit(self.id)



# -*- encoding:utf-8 -*-
# PyQtInspect Logger for Server
import logging

from PyQtInspect._pqi_bundle.pqi_log.base_logger import getLogger


def _get_console_log_level():
    return logging.WARNING


def _get_file_log_level():
    return logging.INFO


logger = getLogger(
    logger_name='pqi_server',
    console_log_level=_get_console_log_level(),
    file_log_level=_get_file_log_level()
)



import sys
import struct
print('Executable: %s' % sys.executable)
import os
def loop_in_thread():
    while True:
        import time
        time.sleep(.5)
        sys.stdout.write('#')
        sys.stdout.flush()

import threading
threading.Thread(target=loop_in_thread).start()


def is_python_64bit():
    return (struct.calcsize('P') == 8)

print('Is 64: %s' % is_python_64bit())

if __name__ == '__main__':
    print('pid:%s' % (os.getpid()))
    i = 0
    while True:
        i += 1
        import time
        time.sleep(.5)
        sys.stdout.write('.')
        sys.stdout.flush()
        if i % 40 == 0:
            sys.stdout.write('\n')
            sys.stdout.flush()



import marshal
import sys

# todo
if sys.version_info < (3, 7):
    BYTE_CODE_OFFSET = 12
elif sys.version_info <= (3, 13):
    BYTE_CODE_OFFSET = 16


def get_magic_number_of_executable():
    import importlib.util
    return importlib.util.MAGIC_NUMBER


def exec_pyc(pyc_file_path: str, globals_=None, locals_=None):
    if globals_ is None:
        globals_ = {}
    globals_.update({
        '__name__': '__main__',
        '__file__': pyc_file_path,
    })
    with open(pyc_file_path, 'rb') as pyc:
        pyc_magic_number = pyc.read(4)
        if pyc_magic_number != get_magic_number_of_executable():
            raise ValueError(f'The magic number of {pyc_file_path} is not equal to the magic number of executable.')

        pyc.seek(BYTE_CODE_OFFSET)
        content = pyc.read()
        code = marshal.loads(content)
        exec(code, globals_, locals_)


def compile_file(file_path):
    import py_compile
    py_compile.compile(file_path, cfile=file_path + 'c')


if __name__ == '__main__':
    exec_pyc('example.pyc')



# -*- encoding:utf-8 -*-

import compileall
import sys
import os
import hashlib

dirname = os.path.dirname


def compile_pqi_module():
    compileall.compile_dir("..", force=True)


def copy_pyc_files_to(dest_dir: str):
    import os
    import shutil
    for root, dirs, files in os.walk(".."):
        for file in files:
            if file.endswith(".pyc"):
                src_file = os.path.join(root, file)
                dest_file = os.path.join(dest_dir, src_file)
                # remove `cpython-xxx`
                dest_file = dest_file.replace("cpython-37.", "")
                # remove `__pycache__`
                dest_file = dest_file.replace("__pycache__\\", "")
                sub_dir = os.path.dirname(dest_file)
                if not os.path.exists(sub_dir):
                    os.makedirs(sub_dir)

                shutil.copy(src_file, dest_file)


def need_compile(output_dir):
    metadata_path = output_dir + "/_PQI_COMPILE_METADATA"
    if not os.path.exists(metadata_path):
        return True
    with open(metadata_path, "r") as f:
        md5 = f.readline().strip()
        if md5 != hashlib.md5(open(sys.executable, 'rb').read()).hexdigest():
            return True
    return False


def write_metadata(output_dir):
    with open(output_dir + "/_PQI_COMPILE_METADATA", "w") as f:
        f.write(hashlib.md5(open(sys.executable, 'rb').read()).hexdigest())


def compile_pqi_module_new(output_dir):
    import py_compile
    module_path = dirname(dirname(os.path.abspath(__file__)))

    for root, dirs, files in os.walk(module_path):
        if '_pqi_compile' in root:  # skip _pqi_compile folder
            continue

        for file in files:
            if file.endswith(".py"):
                src_file = os.path.join(root, file)

                # remove the module path
                src_relative_path = os.path.relpath(src_file, module_path)

                dest_file = os.path.join(output_dir, src_relative_path + "c")  # xxx/xxx.py -> xxx/xxx.pyc
                sub_dir = dirname(dest_file)
                if not os.path.exists(sub_dir):
                    os.makedirs(sub_dir)
                py_compile.compile(src_file, dest_file)


if __name__ == '__main__':
    dstPath = sys.argv[1]
    if need_compile(dstPath):
        print(f'[PyQtInspect] Compile pqi module by {sys.executable}...')
        compile_pqi_module_new(dstPath)
        write_metadata(dstPath)
    else:
        print(f'[PyQtInspect] Not need compile pqi module')



# -*- encoding:utf-8 -*-

"""
Developed by Jianzhang Chen
LICENSE: MIT
"""
import sys

if sys.platform != 'win32':
    raise NotImplementedError('Only support Windows platform')

import threading

from ctypes import POINTER, cast, byref, Structure, WinError, get_last_error, c_int, WinDLL, WINFUNCTYPE
from ctypes.wintypes import WPARAM, LPARAM, HANDLE, DWORD, BOOL, HINSTANCE, UINT, MSG, HHOOK, HWND

__all__ = ['grab']

user32 = WinDLL('user32', use_last_error=True)

HC_ACTION = 0
WH_KEYBOARD_LL = 13

WM_NULL = 0x0000
WM_QUIT = 0x0012

WM_KEYDOWN = 0x0100
WM_KEYUP = 0x0101
WM_SYSKEYDOWN = 0x0104
WM_SYSKEYUP = 0x0105

VK_F8 = 0x77  # 目前只使用F8键


WM_TO_TEXT = {
    WM_KEYDOWN: 'WM_KEYDOWN',
    WM_KEYUP: 'WM_KEYUP',
    WM_SYSKEYDOWN: 'WM_SYSKEYDOWN',
    WM_SYSKEYUP: 'WM_SYSKEYUP',
}

IMAGE_CURSOR = 2
LR_SHARED = 0x00008000
LR_COPYFROMRESOURCE = 0x00004000

SPI_SETCURSORS = 0x0057

ULONG_PTR = WPARAM
LRESULT = LPARAM
LPMSG = POINTER(MSG)
HCURSOR = HANDLE

HOOKPROC = WINFUNCTYPE(LRESULT, c_int, WPARAM, LPARAM)
LowLevelKeyboardProc = HOOKPROC


class KBDLLHOOKSTRUCT(Structure):
    _fields_ = (('vkCode', DWORD),
                ('scanCode', DWORD),
                ('flags', DWORD),
                ('time', DWORD),
                ('dwExtraInfo', ULONG_PTR))


LPKBDLLHOOKSTRUCT = POINTER(KBDLLHOOKSTRUCT)


def errcheck_bool(result, func, args):
    if not result:
        raise WinError(get_last_error())
    return args


# ===================================
#  SetWindowsHookEx
#  https://learn.microsoft.com/zh-cn/windows/win32/api/winuser/nf-winuser-setwindowshookexw
# ===================================
user32.SetWindowsHookExW.errcheck = errcheck_bool
user32.SetWindowsHookExW.restype = HHOOK
user32.SetWindowsHookExW.argtypes = (
    # _In_ idHook
    c_int,
    # _In_ lpfn
    HOOKPROC,
    # _In_ hMod
    HINSTANCE,
    # _In_ dwThreadId
    DWORD,
)

# ===================================
#  PostThreadMessageW
#  https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-postthreadmessagew
# ===================================
user32.PostThreadMessageW.restype = BOOL
user32.PostThreadMessageW.argtypes = (
    # _In_ idThread
    DWORD,
    # _In_ Msg
    UINT,
    # _In_ wParam
    WPARAM,
    # _In_ lParam
    LPARAM,
)

# ===================================
#  UnhookWindowsHookEx
#  https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-unhookwindowshookex
# ===================================
user32.UnhookWindowsHookEx.restype = BOOL
user32.UnhookWindowsHookEx.argtypes = (
    # _In_ hhk
    HHOOK,
)

# ===================================
#  CallNextHookEx
#  https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-callnexthookex
# ===================================
user32.CallNextHookEx.restype = LRESULT
user32.CallNextHookEx.argtypes = (
    # _In_opt_ hhk
    HHOOK,
    # _In_     nCode
    c_int,
    # _In_     wParam
    WPARAM,
    # _In_     lParam
    LPARAM,
)

# ===================================
#  GetMessageW
#  https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-getmessagew
# ===================================
user32.GetMessageW.argtypes = (
    # _Out_    lpMsg
    LPMSG,
    # _In_opt_ hWnd
    HWND,
    # _In_     wMsgFilterMin
    UINT,
    # _In_     wMsgFilterMax
    UINT,
)

# ===================================
#  TranslateMessage
#  https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-translatemessage
# ===================================
user32.TranslateMessage.argtypes = (
    # _In_ lpMsg
    LPMSG,
)

# ===================================
#  DispatchMessageW
#  https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-dispatchmessagew
# ===================================
user32.DispatchMessageW.argtypes = (
    # _In_ lpMsg
    LPMSG,
)


# Whether to print debug messages
_is_debug = False

# The result of the grab
_result = 0

# The target virtual key
_target_vk = 0

SUCCESS_RESULT = 1


def _print_keyboard_msg(wParam, msg):
    msg_id = WM_TO_TEXT.get(wParam, str(wParam))
    msg_to_print = (msg.vkCode, msg.scanCode, msg.flags, msg.time, msg.dwExtraInfo)
    print('{:15s}: {}'.format(msg_id, msg_to_print))


@LowLevelKeyboardProc
def _LLKeyboardProc(nCode, wParam, lParam):
    """ Low-level mouse input event hook procedure. """
    global _result

    if nCode == HC_ACTION:
        msg = cast(lParam, LPKBDLLHOOKSTRUCT)[0]

        if _is_debug:
            _print_keyboard_msg(wParam, msg)

        if wParam == WM_KEYUP and msg.vkCode == _target_vk:
            # Post a WM_NULL message to the current thread to exit the message loop when the grab is finished.
            _result = SUCCESS_RESULT
            user32.PostThreadMessageW(threading.current_thread().ident, WM_NULL, 0, 0)
            return 1
    return user32.CallNextHookEx(None, nCode, wParam, lParam)


class GrabFlag:
    def __init__(self):
        self._stop = False

    def mark_stop(self):
        # may be called by another thread
        self._stop = True

    def is_stop(self):
        return self._stop

    def clear_flag(self):
        self._stop = False


def _msg_loop(flag):
    """ Start a message loop to grab the PID of the window under the cursor. """
    hook = user32.SetWindowsHookExW(WH_KEYBOARD_LL, _LLKeyboardProc, None, 0)
    msg = MSG()

    while _result == 0 and not flag.is_stop():  # Wait until the grab is finished (when _result is not zero).
        bRet = user32.GetMessageW(byref(msg), None, 0, 0)
        if not bRet:
            break
        if bRet == -1:
            raise WinError(get_last_error())
        user32.TranslateMessage(byref(msg))
        user32.DispatchMessageW(byref(msg))

    flag.mark_stop()
    user32.UnhookWindowsHookEx(hook)
    return _result


# region The public API
def grab(virtual_key, flag, callback, _debug=False):
    if not isinstance(flag, GrabFlag):
        raise TypeError('flag must be an instance of GrabFlag')

    global _is_debug, _result, _target_vk
    _is_debug = _debug
    _target_vk = virtual_key

    _result = 0
    _msg_loop(flag)
    if _result == SUCCESS_RESULT:
        callback()
    return


# endregion


if __name__ == '__main__':
    _flag = GrabFlag()
    grab(VK_F8, _flag, callback=lambda: print('<!UP!>'), _debug=False)



_DEFAULT_HOST = '127.0.0.1'
_DEFAULT_PORT = 19394


class PQIServerGUIDataCenter:
    def __init__(self):
        self._serverConfig = {}

    def setServerConfig(self, config: dict):
        self._serverConfig = config

    @property
    def host(self):
        return self._serverConfig.get('host', _DEFAULT_HOST)

    @property
    def port(self):
        return self._serverConfig.get('port', _DEFAULT_PORT)


instance = PQIServerGUIDataCenter()



# -*- encoding:utf-8 -*-
# ==============================================
# Author: Jeza Chen
# Time: 2023/8/18 15:00
# Description: 
# ==============================================


class ArgHandlerWithParam:
    '''
    Handler for some arguments which needs a value
    '''

    def __init__(self, arg_name, convert_val=None, default_val=None):
        self.arg_name = arg_name
        self.arg_v_rep = '--%s' % (arg_name,)
        self.convert_val = convert_val
        self.default_val = default_val

    def to_argv(self, lst, setup):
        v = setup.get(self.arg_name)
        if v is not None and v != self.default_val:
            lst.append(self.arg_v_rep)
            lst.append('%s' % (v,))

    def handle_argv(self, argv, i, setup):
        assert argv[i] == self.arg_v_rep
        del argv[i]

        val = argv[i]
        if self.convert_val:
            val = self.convert_val(val)

        setup[self.arg_name] = val
        del argv[i]


# to_argv输出带等号的,"--arg_name=arg_value"
class ArgHandlerWithParamAndEqualSign(ArgHandlerWithParam):
    def to_argv(self, lst, setup):
        v = setup.get(self.arg_name)
        if v is not None and v != self.default_val:
            lst.append('%s=%s' % (self.arg_v_rep, v))


class ArgHandlerBool:
    '''
    If a given flag is received, mark it as 'True' in setup.
    '''

    def __init__(self, arg_name, default_val=False):
        self.arg_name = arg_name
        self.arg_v_rep = '--%s' % (arg_name,)
        self.default_val = default_val

    def to_argv(self, lst, setup):
        v = setup.get(self.arg_name)
        if v:
            lst.append(self.arg_v_rep)

    def handle_argv(self, argv, i, setup):
        assert argv[i] == self.arg_v_rep
        del argv[i]
        setup[self.arg_name] = True


ACCEPTED_ARG_HANDLERS = [
    ArgHandlerWithParam('port', int, 19394),
    ArgHandlerWithParam('client', default_val='127.0.0.1'),
    ArgHandlerWithParam('stack-max-depth', int, 0),

    ArgHandlerBool('server'),
    ArgHandlerBool('DEBUG_RECORD_SOCKET_READS'),
    ArgHandlerBool('multiprocess'),  # Used by PyDev (creates new connection to ide)
    ArgHandlerBool('module'),
    ArgHandlerBool('help'),
    ArgHandlerBool('show-pqi-stack'),
    ArgHandlerWithParamAndEqualSign('qt-support', default_val='pyqt5'),  # 原来的pydevd不支持子进程带qt参数, 这里加一下
]

ARGV_REP_TO_HANDLER = {}
for handler in ACCEPTED_ARG_HANDLERS:
    ARGV_REP_TO_HANDLER[handler.arg_v_rep] = handler


def get_pydevd_file(executable_path):
    import PyQtInspect.pqi

    f = PyQtInspect.pqi.__file__
    if f.endswith('.pyc'):
        f = f[:-1]
    elif f.endswith('$py.class'):
        f = f[:-len('$py.class')] + '.py'

    return f


def setup_to_argv(executable_path, setup):
    '''
    :param dict setup:
        A dict previously gotten from process_command_line.

    :note: does not handle --file nor --DEBUG.
    '''
    ret = [get_pydevd_file(executable_path)]

    for handler in ACCEPTED_ARG_HANDLERS:
        if handler.arg_name in setup:
            handler.to_argv(ret, setup)
    return ret


def process_command_line(argv):
    """ parses the arguments.
        removes our arguments from the command line """
    setup = {}
    for handler in ACCEPTED_ARG_HANDLERS:
        setup[handler.arg_name] = handler.default_val
    setup['file'] = ''
    setup['show-pqi-stack'] = False

    i = 0
    del argv[0]
    while i < len(argv):
        handler = ARGV_REP_TO_HANDLER.get(argv[i])
        if handler is not None:
            handler.handle_argv(argv, i, setup)

        elif argv[i].startswith('--qt-support'):
            # The --qt-support is special because we want to keep backward compatibility:
            # Previously, just passing '--qt-support' meant that we should use the auto-discovery mode
            # whereas now, if --qt-support is passed, it should be passed as --qt-support=<mode>, where
            # mode can be one of 'auto', 'none', 'pyqt5', 'pyqt4', 'pyside', 'pyside2'.
            if argv[i] == '--qt-support':
                setup['qt-support'] = 'auto'

            elif argv[i].startswith('--qt-support='):
                qt_support = argv[i][len('--qt-support='):]
                qt_support = qt_support.lower()
                valid_modes = ('pyqt5', 'pyqt6', 'pyside2', 'pyside6')  # TODO ONLY SUPPORTS PYQT5/6 AND PYSIDE2/6
                if qt_support not in valid_modes:
                    raise ValueError("qt-support mode invalid: " + qt_support)
                if qt_support == 'none':
                    # On none, actually set an empty string to evaluate to False.
                    setup['qt-support'] = ''
                else:
                    setup['qt-support'] = qt_support
            else:
                raise ValueError("Unexpected definition for qt-support flag: " + argv[i])

            del argv[i]

        elif argv[i] == '--file':
            # --file is special because it's the last one (so, no handler for it).
            del argv[i]
            setup['file'] = argv[i]
            i = len(argv)  # pop out, file is our last argument

        elif argv[i] == '--DEBUG':
            from PyQtInspect.pqi import set_debug
            del argv[i]
            set_debug(setup)

        else:
            raise ValueError("Unexpected option: " + argv[i])
    return setup



# -*- encoding:utf-8 -*-
# resources for pqi_gui

_PQI_GUI_ICON = None


def get_icon():
    global _PQI_GUI_ICON
    if _PQI_GUI_ICON is None:
        from PyQt5 import QtGui
        from . import resources

        _PQI_GUI_ICON = QtGui.QIcon(":/icons/icon.png")
    return _PQI_GUI_ICON



# PQI Structures

import dataclasses


@dataclasses.dataclass
class QWidgetInfo:
    """A dataclass for storing information about a QWidget."""
    # Class name of the QWidget.
    class_name: str

    # The object name of the QWidget.
    object_name: str

    # The id of the QWidget.
    id: int

    # The stack of the QWidget when it was created.
    stacks_when_create: list

    # The size of the QWidget.
    size: tuple

    # The position of the QWidget.
    pos: tuple

    # === PARENTS ===
    # 之所以拆分成两个列表, 是为了减少传输的JSON数据量
    # TODO: 评估下, 使用嵌套dataclass? namedtuple?
    # The parent classes of the QWidget.
    parent_classes: list

    # The id of all parents of the QWidget.
    parent_ids: list

    # The object name of all parents of the QWidget.
    parent_object_names: list

    # The stylesheet of the QWidget.
    stylesheet: str

    # extra data from sender
    extra: dict = dataclasses.field(default_factory=dict)


@dataclasses.dataclass
class QWidgetChildrenInfo:
    """A dataclass for storing information about a QWidget's ancestor."""
    widget_id: int

    child_classes: list

    child_ids: list

    child_object_names: list



# -*- encoding:utf-8 -*-
# ==============================================
# Author: Jeza Chen
# Time: 2023/8/24 17:36
# Description: 
# ==============================================
import pathlib
import sys
import typing

from PyQtInspect._pqi_bundle import pqi_log
from PyQtInspect.pqi_gui.workers.pqy_worker import PQYWorker, DUMMY_WORKER, DummyWorker

pyqt_inspect_module_dir = str(pathlib.Path(__file__).resolve().parent.parent)
if pyqt_inspect_module_dir not in sys.path:
    sys.path.insert(0, pyqt_inspect_module_dir)

import json

from PyQt5 import QtWidgets, QtCore
from PyQtInspect.pqi_gui.attach_window import AttachWindow
from PyQtInspect.pqi_gui.create_stacks_list_widget import CreateStacksListWidget
from PyQtInspect._pqi_bundle.pqi_comm_constants import CMD_WIDGET_INFO, CMD_INSPECT_FINISHED, CMD_EXEC_CODE_ERROR, \
    CMD_EXEC_CODE_RESULT, CMD_CHILDREN_INFO, CMD_QT_PATCH_SUCCESS

import ctypes

from PyQtInspect.pqi_gui.code_window import CodeWindow
from PyQtInspect.pqi_gui.hierarchy_bar import HierarchyBar
from PyQtInspect.pqi_gui.settings_window import SettingWindow
from PyQtInspect.pqi_gui.styles import GLOBAL_STYLESHEET
import PyQtInspect.pqi_gui.data_center as DataCenter
from PyQtInspect.pqi_gui._pqi_res import get_icon
from PyQtInspect.pqi_gui.keyboard_hook_handler import KeyboardHookHandler


# ==== SetupHolder ====
from PyQtInspect._pqi_common.pqi_setup_holder import SetupHolder

SetupHolder.setup = {
    'server': True
}

if sys.platform == "win32":
    myappid = 'jeza.tools.pyqt_inspect.0.0.1alpha2'
    ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)


class BriefLine(QtWidgets.QWidget):
    def __init__(self, parent, key: str, defaultValue: str = ""):
        super().__init__(parent)
        self.setFixedHeight(30)

        self._layout = QtWidgets.QHBoxLayout(self)
        self._layout.setContentsMargins(5, 0, 5, 0)
        self._layout.setSpacing(5)

        self._keyLabel = QtWidgets.QLabel(self)
        self._keyLabel.setText(key)
        self._keyLabel.setAlignment(QtCore.Qt.AlignCenter)
        self._keyLabel.setWordWrap(True)
        self._keyLabel.setMinimumWidth(90)
        self._keyLabel.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)

        self._layout.addWidget(self._keyLabel)

        self._valueLineEdit = QtWidgets.QLineEdit(self)
        self._valueLineEdit.setObjectName("codeStyleLineEdit")
        self._valueLineEdit.setText(defaultValue)
        self._valueLineEdit.setAlignment(QtCore.Qt.AlignCenter)
        self._valueLineEdit.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        self._valueLineEdit.setReadOnly(True)

        self._layout.addWidget(self._valueLineEdit)

    def setValue(self, value: str):
        self._valueLineEdit.setText(value)


class BriefLineWithEditButton(BriefLine):
    sigEditButtonClicked = QtCore.pyqtSignal(str)  # new value

    def __init__(self, parent, key: str = None, defaultValue: str = None, buttonText: str = "Edit"):
        super().__init__(parent, key, defaultValue)

        self._valueLineEdit.setReadOnly(False)

        self._editButton = QtWidgets.QPushButton(self)
        self._editButton.setText(buttonText)
        self._editButton.setFixedHeight(30)
        self._editButton.clicked.connect(lambda: self.sigEditButtonClicked.emit(self._valueLineEdit.text()))

        self._layout.addWidget(self._editButton)


class WidgetBriefWidget(QtWidgets.QWidget):
    sigOpenCodeWindow = QtCore.pyqtSignal()

    def __init__(self, parent):
        super().__init__(parent)
        self._mainLayout = QtWidgets.QVBoxLayout(self)
        self._mainLayout.setContentsMargins(0, 0, 0, 0)
        self._mainLayout.setSpacing(5)

        self._classNameLine = BriefLine(self, "class name")
        self._mainLayout.addWidget(self._classNameLine)

        self._objectNameLine = BriefLine(self, "object name")
        self._mainLayout.addWidget(self._objectNameLine)

        self._sizeLine = BriefLine(self, "size")
        self._mainLayout.addWidget(self._sizeLine)

        self._posLine = BriefLine(self, "pos")
        self._mainLayout.addWidget(self._posLine)

        self._styleSheetLine = BriefLine(self, "stylesheet")
        self._mainLayout.addWidget(self._styleSheetLine)

        self._executionButtonsLayout = QtWidgets.QHBoxLayout()
        self._executionButtonsLayout.setContentsMargins(4, 0, 4, 0)
        self._executionButtonsLayout.setSpacing(5)

        self._execCodeButton = QtWidgets.QPushButton(self)
        self._execCodeButton.setText("Run Code")
        self._execCodeButton.setFixedHeight(30)
        self._execCodeButton.clicked.connect(self.sigOpenCodeWindow)

        self._executionButtonsLayout.addWidget(self._execCodeButton)

        self._mainLayout.addLayout(self._executionButtonsLayout)

        self._mainLayout.addStretch(1)

    def setInfo(self, info):
        self._classNameLine.setValue(info["class_name"])
        objName = info["object_name"]
        self._objectNameLine.setValue(objName)
        width, height = info["size"]
        self._sizeLine.setValue(f"{width}, {height}")
        posX, posY = info["pos"]
        self._posLine.setValue(f"{posX}, {posY}")
        self._styleSheetLine.setValue(info["stylesheet"])

    def clearInfo(self):
        self._classNameLine.setValue("")
        self._objectNameLine.setValue("")
        self._sizeLine.setValue("")
        self._posLine.setValue("")
        self._styleSheetLine.setValue("")


class PQIWindow(QtWidgets.QMainWindow):
    _sigInspectFinished = QtCore.pyqtSignal()
    _sigInspectBegin = QtCore.pyqtSignal()
    _sigInspectDisabled = QtCore.pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("PyQtInspect")
        self.setWindowIcon(get_icon())
        self.resize(700, 1000)

        self._menuBar = QtWidgets.QMenuBar(self)
        self.setMenuBar(self._menuBar)

        self._moreMenu = QtWidgets.QMenu(self._menuBar)
        self._moreMenu.setTitle("More")
        self._menuBar.addMenu(self._moreMenu)

        # ==================== #
        #     Menu Actions     #
        # ==================== #
        # Press F8 to Disable Inspect Action
        self._keyboardHookHandler = KeyboardHookHandler(self)

        self._pressF8ToDisableInspectAction = QtWidgets.QAction(self)
        self._pressF8ToDisableInspectAction.setText("Press F8 to Finish Inspect")
        self._pressF8ToDisableInspectAction.setCheckable(True)
        self._pressF8ToDisableInspectAction.setEnabled(self._keyboardHookHandler.isValid())
        self._pressF8ToDisableInspectAction.setChecked(self._keyboardHookHandler.isValid())  # default
        # --- Connect Signals ---
        # main gui -> keyboard hook handler
        self._sigInspectBegin.connect(self._keyboardHookHandler.onInspectBegin)
        self._sigInspectFinished.connect(self._keyboardHookHandler.onInspectFinished)
        self._sigInspectDisabled.connect(self._keyboardHookHandler.onInspectDisabled)
        self._pressF8ToDisableInspectAction.toggled.connect(self._keyboardHookHandler.setEnable)
        # keyboard hook handler -> main gui
        self._keyboardHookHandler.sigDisableInspectKeyPressed.connect(self._onInspectKeyPressed)

        self._moreMenu.addAction(self._pressF8ToDisableInspectAction)

        # Mock Left Button Down Action
        self._isMockLeftButtonDownAction = QtWidgets.QAction(self)
        self._isMockLeftButtonDownAction.setText("Mock Right Button Down as Left")
        self._isMockLeftButtonDownAction.setCheckable(True)
        self._isMockLeftButtonDownAction.setChecked(True)  # default
        self._moreMenu.addAction(self._isMockLeftButtonDownAction)

        # Attach Action
        self._attachAction = QtWidgets.QAction(self)
        self._attachAction.setText("Attach To Process")
        self._moreMenu.addAction(self._attachAction)
        self._attachAction.triggered.connect(self._onAttachActionTriggered)
        self._attachAction.setEnabled(False)

        self._moreMenu.addSeparator()

        # Setting Action
        self._settingAction = QtWidgets.QAction(self)
        self._settingAction.setText("Settings")
        self._moreMenu.addAction(self._settingAction)
        self._settingAction.triggered.connect(self._openSettingWindow)

        # Child Windows
        self._settingWindow = None
        self._codeWindow = None
        self._attachWindow = None

        self._mainContainer = QtWidgets.QWidget(self)
        self.setCentralWidget(self._mainContainer)
        self._mainLayout = QtWidgets.QVBoxLayout(self._mainContainer)
        self._mainLayout.setContentsMargins(4, 4, 4, 4)
        self._mainLayout.setSpacing(0)

        # ====================
        #    Top Container   #
        # ====================
        self._topContainer = QtWidgets.QWidget(self)
        self._topContainer.setFixedHeight(30)
        self._topContainer.move(0, 0)

        self._topLayout = QtWidgets.QHBoxLayout(self._topContainer)
        self._topLayout.setContentsMargins(0, 0, 0, 0)
        self._topLayout.setSpacing(0)

        self._portLineEdit = QtWidgets.QLineEdit(self._topContainer)
        self._portLineEdit.setFixedHeight(30)
        self._portLineEdit.setText("19394")
        self._portLineEdit.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)

        self._topLayout.addWidget(self._portLineEdit)

        self._serveButton = QtWidgets.QPushButton(self)
        self._serveButton.setText("Serve")
        self._serveButton.setFixedHeight(30)
        self._serveButton.setCheckable(True)
        self._serveButton.clicked.connect(self._onServeButtonToggled)

        self._topLayout.addWidget(self._serveButton)

        self._inspectButton = QtWidgets.QPushButton(self)
        self._inspectButton.setText("Inspect")
        self._inspectButton.setFixedHeight(30)
        self._inspectButton.setCheckable(True)
        self._inspectButton.clicked.connect(self._onInspectButtonClicked)
        self._inspectButton.setEnabled(False)

        self._topLayout.addWidget(self._inspectButton)

        self._mainLayout.addWidget(self._topContainer)

        self._widgetInfoGroupBox = QtWidgets.QGroupBox(self)
        self._widgetInfoGroupBox.setTitle("Widget Brief Info")

        self._widgetInfoGroupBoxLayout = QtWidgets.QVBoxLayout(self._widgetInfoGroupBox)
        self._widgetInfoGroupBoxLayout.setContentsMargins(0, 4, 0, 6)
        self._widgetInfoGroupBoxLayout.setSpacing(0)

        self._widgetBriefWidget = WidgetBriefWidget(self)
        self._widgetBriefWidget.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        self._widgetBriefWidget.sigOpenCodeWindow.connect(self._openCodeWindow)

        self._widgetInfoGroupBoxLayout.addWidget(self._widgetBriefWidget)

        self._mainLayout.addSpacing(3)
        self._mainLayout.addWidget(self._widgetInfoGroupBox)

        # ==================== #
        #     Create Stack     #
        # ==================== #
        self._createStackGroupBox = QtWidgets.QGroupBox(self)
        self._createStackGroupBox.setTitle("Create Stacks")

        self._createStackGroupBoxLayout = QtWidgets.QVBoxLayout(self._createStackGroupBox)
        self._createStackGroupBoxLayout.setContentsMargins(4, 4, 4, 6)
        self._createStackGroupBoxLayout.setSpacing(0)

        self._createStacksListWidget = CreateStacksListWidget(self)
        self._createStacksListWidget.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)

        self._createStackGroupBoxLayout.addWidget(self._createStacksListWidget)

        self._mainLayout.addSpacing(3)
        self._mainLayout.addWidget(self._createStackGroupBox)

        # ==================== #
        #     Hierarchy Bar    #
        # ==================== #
        self._hierarchyBar = HierarchyBar(self)
        self._hierarchyBar.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        self._hierarchyBar.sigAncestorItemHovered.connect(self._onAncestorWidgetItemHighlight)
        self._hierarchyBar.sigAncestorItemChanged.connect(self._onAncestorWidgetItemClicked)
        self._hierarchyBar.sigChildMenuItemHovered.connect(self._onChildWidgetItemHighlight)
        self._hierarchyBar.sigChildMenuItemClicked.connect(self._onChildWidgetItemClicked)
        self._hierarchyBar.sigReqChildWidgetsInfo.connect(self._reqChildWidgetsInfo)
        self._hierarchyBar.sigMouseLeaveBarAndMenu.connect(self._unhighlightPrevWidget)

        self._mainLayout.addSpacing(3)
        self._mainLayout.addWidget(self._hierarchyBar)

        self._worker = None
        self._currDispatcherIdForSelectedWidget = None
        self._currDispatcherIdForHoveredWidget = None  # todo 可能会有多个进程都有选中的情况?

        self._curWidgetId = -1
        self._curHighlightedWidgetId = -1

        self.setStyleSheet(GLOBAL_STYLESHEET)

    # region For Serve Button
    def _onServeButtonToggled(self, checked: bool):
        if checked:
            try:
                port = int(self._portLineEdit.text())
            except ValueError:
                QtWidgets.QMessageBox.warning(self, "PyQtInspect", "Port must be a number")
                self._serveButton.setChecked(False)
                return

            DataCenter.instance.setServerConfig({"port": port})

            self._runWorker()
        else:
            self._askStopWorkerConfirmation()

    def _runWorker(self):
        if self._worker is not None:
            return

        port = DataCenter.instance.port

        self._portLineEdit.setEnabled(False)
        # self._serveButton.setEnabled(False)
        self._inspectButton.setEnabled(True)
        self._attachAction.setEnabled(True)

        self._worker = PQYWorker(None, port)  # The parent of worker must be None!
        self._worker.widgetInfoRecv.connect(self.on_widget_info_recv)
        self._worker.sigNewDispatcher.connect(self.onNewDispatcher)
        self._worker.socketError.connect(self._onWorkerSocketError)
        self._workerThread = QtCore.QThread()

        self._worker.moveToThread(self._workerThread)
        self._workerThread.started.connect(self._worker.run)

        self._workerThread.start()

    def _cleanUpWhenWorkerStopped(self):
        if self._worker is None:
            return

        # clear worker and its thread

        if self._worker is not None:
            self._worker.stop()
            self._worker.deleteLater()
            self._worker = None

        if self._workerThread is not None:
            self._workerThread.quit()
            self._workerThread.wait()
            self._workerThread = None

        # set buttons status to default
        self._portLineEdit.setEnabled(True)
        self._inspectButton.setEnabled(False)

        # set action status to default
        self._attachAction.setEnabled(False)

        # clear ui
        self._widgetBriefWidget.clearInfo()
        self._createStacksListWidget.clearStacks()
        self._hierarchyBar.clearData()

    def _getWorker(self) -> typing.Union[PQYWorker, DummyWorker]:
        if self._worker is None:
            return DUMMY_WORKER
        return self._worker

    def _askStopWorkerConfirmation(self):
        """ Ask the user for confirmation to stop the server. """
        self._serveButton.setChecked(True)  # hold the button checked before user's choice

        reply = QtWidgets.QMessageBox.question(self, "PyQtInspect", "Are you sure to stop serving?",
                                               QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                                               QtWidgets.QMessageBox.No)
        if reply == QtWidgets.QMessageBox.Yes:
            self._disableInspect()  # disable inspect for all dispatchers before stopping the server
            self._cleanUpWhenWorkerStopped()
            self._serveButton.setChecked(False)

    def _onWorkerSocketError(self, msg):
        QtWidgets.QMessageBox.critical(self, "Error", msg)
        self._cleanUpWhenWorkerStopped()
        self._serveButton.setChecked(False)

    # endregion

    def on_widget_info_recv(self, dispatcherId: int, info: dict):
        cmdId = info.get("cmd_id")
        text = info.get("text", "")
        if cmdId == CMD_QT_PATCH_SUCCESS:
            pid = int(text)
            pqi_log.info(f"PyQtInspect: Qt patched successfully, pid: {pid}")

            # If inspection is enabled, enable it for the new process.
            if self._inspectButton.isChecked():
                self._getWorker().sendEnableInspectToDispatcher(
                    dispatcherId,
                    {'mock_left_button_down': self._isMockLeftButtonDownAction.isChecked()}
                )
        elif cmdId == CMD_WIDGET_INFO:
            self._currDispatcherIdForHoveredWidget = dispatcherId
            self.handleWidgetInfoMsg(json.loads(text))
        elif cmdId == CMD_INSPECT_FINISHED:
            self._currDispatcherIdForSelectedWidget = dispatcherId
            self.handle_inspect_finished_msg()
            self.windowHandle().requestActivate()
        elif cmdId == CMD_EXEC_CODE_ERROR:
            errMsg = text
            self._notifyResultToCodeWindow(True, errMsg)
        elif cmdId == CMD_EXEC_CODE_RESULT:
            result = text
            self._notifyResultToCodeWindow(False, result)
        elif cmdId == CMD_CHILDREN_INFO:
            childrenInfoDict = json.loads(text)
            widgetId = childrenInfoDict["widget_id"]
            self._hierarchyBar.setMenuData(widgetId, childrenInfoDict["child_classes"],
                                           childrenInfoDict["child_object_names"],
                                           childrenInfoDict["child_ids"])
        # self._bottomStatusTextBrowser.append(f"recv: {info}")

    def handleWidgetInfoMsg(self, info):
        self._curWidgetId = info["id"]
        self._widgetBriefWidget.setInfo(info)
        self._createStacksListWidget.setStacks(info.get("stacks_when_create", []))

        # set hierarchy
        if info.get("extra", {}).get("from",
                                     "") != "ancestor":  # 如果为"ancestor", 则意味着用户是通过bar来点击回溯获取祖先控件的信息, 此时无需覆盖祖先控件信息
            classes = [*reversed(info["parent_classes"]), info["class_name"]]
            objNames = [*reversed(info["parent_object_names"]), info["object_name"]]
            ids = [*reversed(info["parent_ids"]), info["id"]]
            self._hierarchyBar.setData(classes, objNames, ids)

    def handle_inspect_finished_msg(self):
        self._inspectButton.setChecked(False)
        self._getWorker().sendDisableInspect()  # disable inspect for all dispatchers
        self._sigInspectFinished.emit()

    def onNewDispatcher(self, dispatcher):
        dispatcher.sigMsg.connect(self.on_widget_info_recv)
        dispatcher.registerMainUIReady()

    def _enableInspect(self):
        worker = self._getWorker()
        if not worker:
            return
        worker.sendEnableInspect({'mock_left_button_down': self._isMockLeftButtonDownAction.isChecked()})
        self._sigInspectBegin.emit()

    def _disableInspect(self):
        self._getWorker().sendDisableInspect()
        self._currDispatcherIdForSelectedWidget = None
        self._sigInspectDisabled.emit()

    def _onInspectButtonClicked(self, checked: bool):
        if not self._getWorker():
            return
        if checked:
            self._enableInspect()
        else:
            self._disableInspect()

    def _notifyExecCodeInSelectedWidget(self, code: str):
        worker = self._getWorker()
        if not worker or self._currDispatcherIdForSelectedWidget is None:
            return

        worker.sendExecCodeEvent(self._currDispatcherIdForSelectedWidget, code)

    def _onAncestorWidgetItemHighlight(self, widgetId: str):
        worker = self._getWorker()
        if worker is None or self._currDispatcherIdForSelectedWidget is None:
            return

        widgetId = int(widgetId)
        # unhighlight prev widget, and highlight current widget
        self._unhighlightPrevWidget()
        worker.sendHighlightWidgetEvent(self._currDispatcherIdForSelectedWidget, widgetId, True)
        self._curHighlightedWidgetId = widgetId

    def _onAncestorWidgetItemClicked(self, widgetId: str):
        worker = self._getWorker()
        if not worker or self._currDispatcherIdForSelectedWidget is None:
            return

        widgetId = int(widgetId)

        self._worker.sendSelectWidgetEvent(self._currDispatcherIdForSelectedWidget, widgetId)
        self._unhighlightPrevWidget()
        self._worker.sendRequestWidgetInfoEvent(self._currDispatcherIdForSelectedWidget, widgetId, {
            "from": "ancestor"
        })  # 通过bar来点击回溯获取祖先控件的信息, 带上from字段, 避免覆盖祖先控件信息(即点击导航条前面的类后, 该类后面的类全都无了, 因为此时显示的是该类的祖先控件信息)
        self._worker.sendRequestChildrenInfoEvent(self._currDispatcherIdForSelectedWidget, widgetId)  # todo 会不会有时序问题

    def _onChildWidgetItemHighlight(self, widgetId: str):
        worker = self._getWorker()
        if worker is None or self._currDispatcherIdForSelectedWidget is None:
            return

        widgetId = int(widgetId)
        # unhighlight prev widget, and highlight current widget
        self._unhighlightPrevWidget()
        worker.sendHighlightWidgetEvent(self._currDispatcherIdForSelectedWidget, widgetId, True)
        self._curHighlightedWidgetId = widgetId

    def _onChildWidgetItemClicked(self, widgetId: str):
        worker = self._getWorker()
        if worker is None or self._currDispatcherIdForSelectedWidget is None:
            return

        widgetId = int(widgetId)
        worker.sendSelectWidgetEvent(self._currDispatcherIdForSelectedWidget, widgetId)
        self._unhighlightPrevWidget()
        worker.sendRequestWidgetInfoEvent(self._currDispatcherIdForSelectedWidget, widgetId)
        worker.sendRequestChildrenInfoEvent(self._currDispatcherIdForSelectedWidget, widgetId)

    def _openSettingWindow(self):
        if self._settingWindow is None:
            self._settingWindow = SettingWindow(self)
        self._settingWindow.show()

    def _openCodeWindow(self):
        if self._codeWindow is None:
            self._codeWindow = CodeWindow(self)
            self._codeWindow.sigExecCode.connect(self._notifyExecCodeInSelectedWidget)
        self._codeWindow.show()

    def _notifyResultToCodeWindow(self, isErr: bool, result: str):
        if self._codeWindow is None:
            return
        self._codeWindow.notifyResult(isErr, result)

    def _reqChildWidgetsInfo(self, widgetIdStr: str):
        worker = self._getWorker()
        if worker is not None and self._currDispatcherIdForSelectedWidget is not None:
            widgetId = int(widgetIdStr)
            worker.sendRequestChildrenInfoEvent(self._currDispatcherIdForSelectedWidget, widgetId)

    def _unhighlightPrevWidget(self):
        # todo 貌似不用这样了现在, pqi可以保证只有一个widget能高亮
        worker = self._getWorker()
        conditions_met = (
            bool(worker),
            self._currDispatcherIdForSelectedWidget is not None,
            self._curHighlightedWidgetId != -1
        )

        if all(conditions_met):
            worker.sendHighlightWidgetEvent(self._currDispatcherIdForSelectedWidget,
                                            self._curHighlightedWidgetId,
                                            False)
            self._curHighlightedWidgetId = -1

    def _onAttachActionTriggered(self):
        if self._attachWindow is None:
            self._attachWindow = AttachWindow(self)
        self._attachWindow.show()

    def _onInspectKeyPressed(self):
        """ 当停止inspect热键按下后, 停止inspect """
        self._inspectButton.setChecked(False)
        self._finishInspectWhenKeyPress()

    def _finishInspectWhenKeyPress(self):
        self._disableInspect()
        # override self._currDispatcherIdForSelectedWidget
        # todo 看看有没有更好的办法
        self._currDispatcherIdForSelectedWidget = self._currDispatcherIdForHoveredWidget

        _worker = self._getWorker()
        # notify the client to select the widget (the widget is marked as inspected only before)
        _worker.sendSelectWidgetEvent(self._currDispatcherIdForSelectedWidget, self._curWidgetId)
        # we must set the highlight status of this widget to false, because it is hovered before
        _worker.sendHighlightWidgetEvent(self._currDispatcherIdForSelectedWidget, self._curWidgetId, False)

    # endregion

    def closeEvent(self, a0):
        self.cleanUp()

    def cleanUp(self):
        self._disableInspect()


def _set_debug():
    import logging
    from PyQtInspect._pqi_bundle.pqi_contants import DebugInfoHolder

    # SetupHolder.setup
    SetupHolder.setup.update({
        'DEBUG_RECORD_SOCKET_READS': True,
        'LOG_TO_FILE_LEVEL': logging.DEBUG,
        'LOG_TO_CONSOLE_LEVEL': logging.DEBUG
    })
    # DebugInfoHolder (for logging)
    DebugInfoHolder.DEBUG_RECORD_SOCKET_READS = SetupHolder.setup.get(
        'DEBUG_RECORD_SOCKET_READS',
        DebugInfoHolder.DEBUG_RECORD_SOCKET_READS)
    DebugInfoHolder.LOG_TO_FILE_LEVEL = SetupHolder.setup.get('LOG_TO_FILE_LEVEL', DebugInfoHolder.LOG_TO_FILE_LEVEL)
    DebugInfoHolder.LOG_TO_CONSOLE_LEVEL = SetupHolder.setup.get('LOG_TO_CONSOLE_LEVEL',
                                                                 DebugInfoHolder.LOG_TO_CONSOLE_LEVEL)


def main():
    import sys

    # Check if the debug option is set.
    debug_option = '--debug'
    if debug_option in sys.argv:
        sys.argv.remove(debug_option)
        _set_debug()

    # open high-resolution support
    QtWidgets.QApplication.setHighDpiScaleFactorRoundingPolicy(
        QtCore.Qt.HighDpiScaleFactorRoundingPolicy.PassThrough
    )
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps)

    app = QtWidgets.QApplication(sys.argv)
    window = PQIWindow()
    window.show()
    sys.exit(app.exec())


if __name__ == '__main__':
    main()



#!~/.wine/drive_c/Python25/python.exe
# -*- coding: utf-8 -*-

# Copyright (c) 2009-2014, Mario Vilas
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice,this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the copyright holder nor the names of its
#       contributors may be used to endorse or promote products derived from
#       this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
Windows application debugging engine for Python.

by Mario Vilas (mvilas at gmail.com)

Project: U{http://sourceforge.net/projects/winappdbg/}

Web:     U{http://winappdbg.sourceforge.net/}

Blog:    U{http://breakingcode.wordpress.com}

@group Debugging:
    Debug, EventHandler, EventSift, DebugLog

@group Instrumentation:
    System, Process, Thread, Module, Window, Registry

@group Disassemblers:
    Disassembler,
    BeaEngine, DistormEngine, PyDasmEngine

@group Crash reporting:
    Crash, CrashDump, CrashDAO, CrashDictionary

@group Memory search:
    Search,
    Pattern,
    BytePattern,
    TextPattern,
    RegExpPattern,
    HexPattern

@group Debug events:
    Event,
    NoEvent,
    CreateProcessEvent,
    CreateThreadEvent,
    ExitProcessEvent,
    ExitThreadEvent,
    LoadDLLEvent,
    UnloadDLLEvent,
    OutputDebugStringEvent,
    RIPEvent,
    ExceptionEvent

@group Win32 API wrappers:
    win32, Handle, ProcessHandle, ThreadHandle, FileHandle

@group Helpers:
    HexInput, HexOutput, HexDump, Color, Table, Logger,
    PathOperations,
    MemoryAddresses,
    CustomAddressIterator,
    DataAddressIterator,
    ImageAddressIterator,
    MappedAddressIterator,
    ExecutableAddressIterator,
    ReadableAddressIterator,
    WriteableAddressIterator,
    ExecutableAndWriteableAddressIterator,
    DebugRegister,
    Regenerator

@group Warnings:
    MixedBitsWarning, BreakpointWarning, BreakpointCallbackWarning,
    EventCallbackWarning, DebugSymbolsWarning, CrashWarning

@group Deprecated classes:
    CrashContainer, CrashTable, CrashTableMSSQL,
    VolatileCrashContainer, DummyCrashContainer

@type version_number: float
@var  version_number: This WinAppDbg major and minor version,
    as a floating point number. Use this for compatibility checking.

@type version: str
@var  version: This WinAppDbg release version,
    as a printable string. Use this to show to the user.

@undocumented: plugins
"""

__revision__ = "$Id$"

# List of all public symbols
__all__ =   [
                # Library version
                'version',
                'version_number',

                # from breakpoint import *
##                'Breakpoint',
##                'CodeBreakpoint',
##                'PageBreakpoint',
##                'HardwareBreakpoint',
##                'Hook',
##                'ApiHook',
##                'BufferWatch',
                'BreakpointWarning',
                'BreakpointCallbackWarning',

                # from crash import *
                'Crash',
                'CrashWarning',
                'CrashDictionary',
                'CrashContainer',
                'CrashTable',
                'CrashTableMSSQL',
                'VolatileCrashContainer',
                'DummyCrashContainer',

                # from debug import *
                'Debug',
                'MixedBitsWarning',

                # from disasm import *
                'Disassembler',
                'BeaEngine',
                'DistormEngine',
                'PyDasmEngine',

                # from event import *
                'EventHandler',
                'EventSift',
##                'EventFactory',
##                'EventDispatcher',
                'EventCallbackWarning',
                'Event',
##                'NoEvent',
                'CreateProcessEvent',
                'CreateThreadEvent',
                'ExitProcessEvent',
                'ExitThreadEvent',
                'LoadDLLEvent',
                'UnloadDLLEvent',
                'OutputDebugStringEvent',
                'RIPEvent',
                'ExceptionEvent',

                # from interactive import *
##                'ConsoleDebugger',

                # from module import *
                'Module',
                'DebugSymbolsWarning',

                # from process import *
                'Process',

                # from system import *
                'System',

                # from search import *
                'Search',
                'Pattern',
                'BytePattern',
                'TextPattern',
                'RegExpPattern',
                'HexPattern',

                # from registry import *
                'Registry',

                # from textio import *
                'HexDump',
                'HexInput',
                'HexOutput',
                'Color',
                'Table',
                'CrashDump',
                'DebugLog',
                'Logger',

                # from thread import *
                'Thread',

                # from util import *
                'PathOperations',
                'MemoryAddresses',
                'CustomAddressIterator',
                'DataAddressIterator',
                'ImageAddressIterator',
                'MappedAddressIterator',
                'ExecutableAddressIterator',
                'ReadableAddressIterator',
                'WriteableAddressIterator',
                'ExecutableAndWriteableAddressIterator',
                'DebugRegister',

                # from window import *
                'Window',

                # import win32
                'win32',

                # from win32 import Handle, ProcessHandle, ThreadHandle, FileHandle
                'Handle',
                'ProcessHandle',
                'ThreadHandle',
                'FileHandle',
            ]

# Import all public symbols
from winappdbg.breakpoint import *
from winappdbg.crash import *
from winappdbg.debug import *
from winappdbg.disasm import *
from winappdbg.event import *
from winappdbg.interactive import *
from winappdbg.module import *
from winappdbg.process import *
from winappdbg.registry import *
from winappdbg.system import *
from winappdbg.search import *
from winappdbg.textio import *
from winappdbg.thread import *
from winappdbg.util import *
from winappdbg.window import *

import winappdbg.win32
from winappdbg.win32 import Handle, ProcessHandle, ThreadHandle, FileHandle

try:
    from sql import *
    __all__.append('CrashDAO')
except ImportError:
    import warnings
    warnings.warn("No SQL database support present (missing dependencies?)",
                  ImportWarning)

# Library version
version_number = 1.5
version = "Version %s" % version_number



# -*- encoding:utf-8 -*-
# ==============================================
# Author: Jeza Chen
# Time: 2023/9/7 20:49
# Description: 
# ==============================================
from PyQt5 import QtWidgets, QtGui, QtCore
from io import StringIO
from contextlib import redirect_stdout

from PyQtInspect.pqi_gui._pqi_res import get_icon
from PyQtInspect.pqi_gui.syntax import PythonHighlighter

CODE_TEXT_EDIT_STYLESHEET = """
QTextEdit#CodeTextEdit {
    font: 14px "Consolas";
}

QTextEdit#ResultTextBrowser {
    font: 14px "Consolas";
}
"""

PROMPT = """# In the following code, self refers to the selected widget. 
# For example, if you want to resize the widget, you can type self.setFixedSize(10, 10).
"""


class CodeTextEdit(QtWidgets.QTextEdit):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setObjectName("CodeTextEdit")

    def event(self, e: QtCore.QEvent) -> bool:
        if e.type() == QtCore.QEvent.KeyPress:
            if e.key() == QtCore.Qt.Key_Tab:
                self.insertPlainText("    ")
                return True
        return super().event(e)


class ResultTextBrowser(QtWidgets.QTextBrowser):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setObjectName("ResultTextBrowser")
        self.setReadOnly(True)


class CodeWindow(QtWidgets.QDialog):
    sigExecCode = QtCore.pyqtSignal(str)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Exec Code For Selected Widget")
        self.setWindowIcon(get_icon())
        self.resize(800, 500)

        self._mainLayout = QtWidgets.QVBoxLayout(self)
        self._mainLayout.setContentsMargins(0, 0, 0, 0)
        self._mainLayout.setSpacing(5)
        self._mainLayout.addSpacing(4)

        self._codeTextEdit = CodeTextEdit(self)
        self._codeTextEdit.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self._codeTextEdit.setText(PROMPT)
        self._codeTextEdit.setWordWrapMode(QtGui.QTextOption.NoWrap)

        self._highlight = PythonHighlighter(self._codeTextEdit.document())
        self._mainLayout.addWidget(self._codeTextEdit)

        self._resultTextBrowser = ResultTextBrowser(self)
        self._resultTextBrowser.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self._mainLayout.addWidget(self._resultTextBrowser)

        self._buttonLayout = QtWidgets.QHBoxLayout()
        self._buttonLayout.setContentsMargins(0, 0, 0, 0)
        self._buttonLayout.setSpacing(5)

        self._runButton = QtWidgets.QPushButton(self)
        self._runButton.setFixedSize(100, 40)
        self._runButton.setText("Run")
        self._runButton.clicked.connect(self._runCode)
        self._buttonLayout.addWidget(self._runButton)

        self._cancelButton = QtWidgets.QPushButton(self)
        self._cancelButton.setFixedSize(100, 40)
        self._cancelButton.setText("Close")
        self._cancelButton.clicked.connect(self.close)
        self._buttonLayout.addWidget(self._cancelButton)

        self._mainLayout.addLayout(self._buttonLayout)

        self._mainLayout.addSpacing(4)

        self.setStyleSheet(CODE_TEXT_EDIT_STYLESHEET)

    def _runCode(self):
        code = self._codeTextEdit.toPlainText()
        self.sigExecCode.emit(code)

    def notifyResult(self, isErr: bool, result: str):
        """ if isErr is True, print the result with the color red, else green. """
        if isErr:
            self._resultTextBrowser.setTextColor(QtGui.QColor(255, 0, 0))
        else:
            self._resultTextBrowser.setTextColor(QtGui.QColor(0, 0, 0))
        self._resultTextBrowser.setText(result)


if __name__ == '__main__':
    import sys

    app = QtWidgets.QApplication(sys.argv)
    window = CodeWindow()
    window.show()
    sys.exit(app.exec())



# -*- encoding:utf-8 -*-
import sys
from PyQt5 import QtCore
import traceback
from socket import socket, AF_INET, SOCK_STREAM, SOL_SOCKET, SO_REUSEADDR, SHUT_RDWR

from PyQtInspect.pqi_gui.workers.dispatcher import Dispatcher


class PQYWorker(QtCore.QObject):
    start = QtCore.pyqtSignal()
    widgetInfoRecv = QtCore.pyqtSignal(dict)
    sigNewDispatcher = QtCore.pyqtSignal(Dispatcher)
    socketError = QtCore.pyqtSignal(str)

    def __init__(self, parent, port):
        super().__init__(parent)
        self.port = port

        self.dispatchers = []
        self.idToDispatcher = {}

        self.start.connect(self.run)

        self._isServing = False
        self._socket = None

    def run(self):
        self._isServing = True
        self._socket = socket(AF_INET, SOCK_STREAM)
        self._socket.settimeout(None)

        try:
            from socket import SO_REUSEPORT
            self._socket.setsockopt(SOL_SOCKET, SO_REUSEPORT, 1)
        except ImportError:
            self._socket.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        try:
            self._socket.bind(('', self.port))
            self._socket.listen(1)

            dispatcherId = 0

            while self._isServing:
                newSock, _addr = self._socket.accept()
                # 新建个线程来处理
                dispatcher = Dispatcher(None, newSock, dispatcherId)
                dispatcher.sigDelete.connect(self._onDispatcherDelete)
                self.dispatchers.append(dispatcher)
                self.idToDispatcher[dispatcherId] = dispatcher

                self.sigNewDispatcher.emit(dispatcher)
                dispatcher.start()
                dispatcherId += 1

        except Exception as e:
            if not self._isServing or getattr(e, 'errno') == 10038:
                return  # Socket closed.

            sys.stderr.write("Could not bind to port: %s\n" % (self.port,))
            sys.stderr.flush()
            traceback.print_exc()
            self.socketError.emit(str(e))

    def stop(self):
        self._isServing = False
        for dispatcher in self.dispatchers:
            dispatcher.close()

        if self._socket:
            # ---
            # For linux, we need to shut down the socket before close it.
            # Otherwise, the socket will be in TIME_WAIT state and block program.
            # ---
            try:
                self._socket.shutdown(SHUT_RDWR)
            except:
                # ---
                # Fixed 20240820: Ignore the exception when shutdown the socket in macOS.
                # ---
                pass
            self._socket.close()

    def onMsg(self, info: dict):
        self.widgetInfoRecv.emit(info)

    def sendEnableInspect(self, extra: dict):
        for dispatcher in self.dispatchers:
            dispatcher.sendEnableInspect(extra)

    def sendEnableInspectToDispatcher(self, dispatcherId: int, extra: dict):
        dispatcher = self.idToDispatcher.get(dispatcherId)
        if dispatcher:
            dispatcher.sendEnableInspect(extra)

    def sendDisableInspect(self):
        for dispatcher in self.dispatchers:
            dispatcher.sendDisableInspect()

    def sendExecCodeEvent(self, dispatcherId: int, code: str):
        dispatcher = self.idToDispatcher.get(dispatcherId)
        if dispatcher:
            dispatcher.sendExecCodeEvent(code)

    def sendHighlightWidgetEvent(self, dispatcherId: int, widgetId: int, isHighlight: bool):
        dispatcher = self.idToDispatcher.get(dispatcherId)
        if dispatcher:
            dispatcher.sendHighlightWidgetEvent(widgetId, isHighlight)

    def sendSelectWidgetEvent(self, dispatcherId: int, widgetId: int):
        dispatcher = self.idToDispatcher.get(dispatcherId)
        if dispatcher:
            dispatcher.sendSelectWidgetEvent(widgetId)

    def sendRequestWidgetInfoEvent(self, dispatcherId: int, widgetId: int, extra: dict = None):
        dispatcher = self.idToDispatcher.get(dispatcherId)
        if dispatcher:
            dispatcher.sendRequestWidgetInfoEvent(widgetId, extra)

    def sendRequestChildrenInfoEvent(self, dispatcherId: int, widgetId: int):
        dispatcher = self.idToDispatcher.get(dispatcherId)
        if dispatcher:
            dispatcher.sendRequestChildrenInfoEvent(widgetId)

    def _onDispatcherDelete(self, id: int):
        dispatcher = self.idToDispatcher.pop(id)
        self.dispatchers.remove(dispatcher)
        dispatcher.close()
        dispatcher.deleteLater()


class DummyWorker:
    def __getattr__(self, item):
        return lambda *args, **kwargs: None

    def __bool__(self):
        return False


DUMMY_WORKER = DummyWorker()



# -*- encoding:utf-8 -*-
import pathlib

__all__ = [
    'is_relative_to',
    'find_pqi_module_path',
    'find_compile_pqi_tool',
]

_PQI_COMPILE_SUBDIR = '_pqi_compile'
_COMPILE_PQI_TOOL_PY = 'compile_pqi.py'
_PQI_ENTRY_FILE_NAME = 'pqi.py'


# === COMMON UTILS ===

def is_relative_to(path, parent):
    """ check if path is relative to parent """
    path = pathlib.Path(path)
    parent = pathlib.Path(parent)
    try:
        return path.is_relative_to(parent)
    except AttributeError:
        return str(path).startswith(str(parent))


# === FOR PQI SELF ===
_PQI_MODULE_PATH_CACHE = None


def find_pqi_module_path():
    """ get the absolute path of pqi.py """
    global _PQI_MODULE_PATH_CACHE

    if _PQI_MODULE_PATH_CACHE is not None:
        return _PQI_MODULE_PATH_CACHE

    path = pathlib.Path(__file__).parent.parent
    if not path.exists():
        raise FileNotFoundError(f'Cant find PyQtInspect module at {path}')
    result = str(path).replace('\\', '/')
    _PQI_MODULE_PATH_CACHE = result
    return result


# === FOR COMPILE ===
_COMPILE_PQI_TOOL_PATH_CACHE = None


def find_compile_pqi_tool():
    """ get the absolute path of compile_pqi.py """
    global _COMPILE_PQI_TOOL_PATH_CACHE

    if _COMPILE_PQI_TOOL_PATH_CACHE is not None:
        return _COMPILE_PQI_TOOL_PATH_CACHE

    path = pathlib.Path(__file__).parent.parent / _PQI_COMPILE_SUBDIR / _COMPILE_PQI_TOOL_PY
    if not path.exists():
        raise FileNotFoundError(f'Cant find {_COMPILE_PQI_TOOL_PY} at {path}')
    result = str(path).replace('\\', '/')
    _COMPILE_PQI_TOOL_PATH_CACHE = result
    return result



# -*- encoding:utf-8 -*-
# Thanks to Charles Machalow
# https://gist.github.com/csm10495/39dde7add5f1b1e73c4e8299f5df1116

import sys
import inspect
import typing


class FrameInfo(typing.NamedTuple):
    """ The information of a frame in the stack. """
    filename: str
    line_no: int
    func_name: str


def getStackFrame(useGetFrame=True) -> typing.List[FrameInfo]:
    """
    Brief:
        Gets a stack frame with the passed in num on the stack.
            If useGetFrame, uses sys._getframe (implementation detail of Cython)
                Otherwise or if sys._getframe is missing, uses inspect.stack() (which is really slow).
    Update:
        - 20240820: We CANNOT Store the raw frame object outputted by `sys.getframe()`
            because it will cause memory leak. We should store the information we need.
    """
    # Not all versions of python have the sys._getframe() method.
    # All should have inspect, though it is really slow
    if useGetFrame and hasattr(sys, '_getframe'):
        frame = sys._getframe(0)
        frames = [
            FrameInfo(
                filename=frame.f_code.co_filename,
                line_no=frame.f_lineno,
                func_name=frame.f_code.co_name
            )
        ]  # 需要在创建期间获取stack就捕获行号, 否则后续的f_lineno会走到最后一行去

        while frame.f_back is not None:
            frames.append(
                FrameInfo(
                    filename=frame.f_back.f_code.co_filename,
                    line_no=frame.f_back.f_lineno,
                    func_name=frame.f_back.f_code.co_name
                )
            )
            frame = frame.f_back

        return frames
    return [FrameInfo(*frame[1:4]) for frame in inspect.stack()]


# -*- encoding:utf-8 -*-
# ==============================================
# Author: Jeza Chen
# Time: 2023/8/18 14:52
# Description: 
# ==============================================
import sys
import os
import time
import pathlib

pyqt_inspect_module_dir = str(pathlib.Path(__file__).resolve().parent.parent)
if pyqt_inspect_module_dir not in sys.path:
    sys.path.insert(0, pyqt_inspect_module_dir)

from PyQtInspect._pqi_bundle._pqi_monkey_qt_helpers import filter_trace_stack
from PyQtInspect._pqi_bundle.pqi_comm_constants import CMD_PROCESS_CREATED, CMD_QT_PATCH_SUCCESS
from PyQtInspect._pqi_bundle.pqi_qt_tools import exec_code_in_widget, get_parent_info, get_widget_size, get_widget_pos, \
    get_stylesheet, get_children_info, set_widget_highlight, get_widget_object_name, is_wrapped_pointer_valid
from PyQtInspect._pqi_imps._pqi_saved_modules import threading, thread
from PyQtInspect._pqi_bundle.pqi_contants import get_current_thread_id, SHOW_DEBUG_INFO_ENV, DebugInfoHolder, IS_WINDOWS
from PyQtInspect._pqi_bundle.pqi_comm import PyDBDaemonThread, ReaderThread, get_global_debugger, set_global_debugger, \
    WriterThread, start_client, start_server, CommunicationRole, NetCommand, NetCommandFactory
from PyQtInspect._pqi_bundle.pqi_typing import OptionalDict
from PyQtInspect._pqi_bundle.pqi_structures import QWidgetInfo, QWidgetChildrenInfo
from PyQtInspect._pqi_bundle import pqi_log

import traceback

threadingCurrentThread = threading.current_thread


def enable_qt_support(qt_support_mode):
    import PyQtInspect._pqi_bundle.pqi_monkey_qt as monkey_qt
    monkey_qt.patch_qt(qt_support_mode)


def get_fullname(mod_name):
    import pkgutil

    try:
        loader = pkgutil.get_loader(mod_name)
    except:
        return None
    if loader is not None:
        for attr in ("get_filename", "_get_filename"):
            meth = getattr(loader, attr, None)
            if meth is not None:
                return meth(mod_name)
    return None


def get_package_dir(mod_name):
    for path in sys.path:
        mod_path = os.path.join(path, mod_name.replace('.', '/'))
        if os.path.isdir(mod_path):
            return mod_path
    return None


def save_main_module(file, module_name):
    # patch provided by: Scott Schlesier - when script is run, it does not
    # use globals from pydevd:
    # This will prevent the pydevd script from contaminating the namespace for the script to be debugged
    # pretend pydevd is not the main module, and
    # convince the file to be debugged that it was loaded as main
    sys.modules[module_name] = sys.modules['__main__']
    sys.modules[module_name].__name__ = module_name

    try:
        from importlib.machinery import ModuleSpec
        from importlib.util import module_from_spec
        m = module_from_spec(ModuleSpec('__main__', loader=None))
    except:
        # A fallback for Python <= 3.4
        from imp import new_module
        m = new_module('__main__')

    sys.modules['__main__'] = m
    orig_module = sys.modules[module_name]
    for attr in ['__loader__', '__spec__']:
        if hasattr(orig_module, attr):
            orig_attr = getattr(orig_module, attr)
            setattr(m, attr, orig_attr)
    m.__file__ = file

    return m


def execfile(file, glob=None, loc=None):
    if glob is None:
        import sys
        glob = sys._getframe().f_back.f_globals
    if loc is None:
        loc = glob

    # It seems that the best way is using tokenize.open(): http://code.activestate.com/lists/python-dev/131251/
    import tokenize
    stream = tokenize.open(file)  # @UndefinedVariable
    try:
        contents = stream.read()
    except:
        from PyQtInspect._pqi_imps._pqi_execfile import exec_pyc
        import sys
        pqi_log.debug(f'exec_pyc: {file}, sys.argv: {sys.argv}')
        exec_pyc(file, glob, loc)
        return
    finally:
        stream.close()

    # execute the script (note: it's important to compile first to have the filename set in debug mode)
    exec(compile(contents + "\n", file, 'exec'), glob, loc)


# =======================================================================================================================
# SetupHolder
# =======================================================================================================================
from PyQtInspect._pqi_common.pqi_setup_holder import SetupHolder


class TrackedLock(object):
    """The lock that tracks if it has been acquired by the current thread
    """

    def __init__(self):
        self._lock = thread.allocate_lock()
        # thread-local storage
        self._tls = threading.local()
        self._tls.is_lock_acquired = False

    def acquire(self):
        self._lock.acquire()
        self._tls.is_lock_acquired = True

    def release(self):
        self._lock.release()
        self._tls.is_lock_acquired = False

    def __enter__(self):
        self.acquire()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()

    def is_acquired_by_current_thread(self):
        return self._tls.is_lock_acquired


connected = False


class PyDB(object):
    """ Main debugging class
    Lots of stuff going on here:

    PyDB starts two threads on startup that connect to remote debugger (RDB)
    The threads continuously read & write commands to RDB.
    PyDB communicates with these threads through command queues.
       Every RDB command is processed by calling process_net_command.
       Every PyDB net command is sent to the net by posting NetCommand to WriterThread queue

       Some commands need to be executed on the right thread (suspend/resume & friends)
       These are placed on the internal command queue.
    """

    _RECONNECT_DELAY = 1
    _RECONNECT_TRIES = 10 if IS_WINDOWS else 100

    def __init__(self, set_as_global=True):
        if set_as_global:
            set_global_debugger(self)
            # pydevd_tracing.replace_sys_set_trace_func()

        self._last_host = None
        self._last_port = None

        self.reader = None
        self.writer = None
        self.cmd_factory = NetCommandFactory()
        # self._cmd_queue = defaultdict(_queue.Queue)  # Key is thread id or '*', value is Queue

        self.breakpoints = {}

        self.ready_to_run = True

        self._finish_debugging_session = False

        # the role PyDB plays in the communication with IDE
        self.communication_role = None

        self.inspect_enabled = False
        self._inspect_extra_data = {}
        self._selected_widget = None

        # Mapping from QWidget object's ID to QWidget object
        # The reason for using dict instead of WeakValueDictionary,
        # is that for some internal child controls (such as QSpinbox.lineEdit()),
        # there are no references at the Python level.
        # This results in the wrapper objects being recycled by gc, and subsequently, these controls can't be found.
        # Using dict can prolong the lifecycle of these wrappers,
        # but each time it's accessed, it needs to check whether it is valid, if it is not, it needs to be removed.
        self._id_to_widget = {}
        self.global_event_filter = None
        self.global_native_event_filter = None

    def _try_reconnect(self):
        """
        Attempts to reconnect to the last host and port used for connection.
        Retries up to 10 times with a delay between retries. If a successful
        reconnection is made, it prints a message indicating the reconnection.
        Returns True if reconnection was successful, False otherwise.
        """
        retry_count = 0
        while retry_count <= self._RECONNECT_TRIES:
            try:
                self.connect(self._last_host, self._last_port)
                pqi_log.info("Reconnected to %s:%s" % (self._last_host, self._last_port))
                return True  # success
            except:
                retry_count += 1
                if not IS_WINDOWS:
                    # ---
                    # For non-Windows platforms, we need to wait a bit longer between retries
                    # because in Windows, connect() will block until the connection is established or times out.
                    # ---
                    time.sleep(self._RECONNECT_DELAY)
        return False

    def finish_debugging_session(self):
        if not self._try_reconnect():
            self._finish_debugging_session = True

    def initialize_network(self, sock):
        sock.settimeout(None)

        self.writer = WriterThread(sock)
        self.reader = ReaderThread(sock)
        self.writer.start()
        self.reader.start()

    def connect(self, host, port):
        self._last_host, self._last_port = host, port
        if host:
            self.communication_role = CommunicationRole.CLIENT
            s = start_client(host, port)
        else:
            self.communication_role = CommunicationRole.SERVER
            s = start_server(port)

        self.initialize_network(s)
        if host:
            self.send_process_created_message()

    def check_output_redirect(self):
        pass

    def send_process_created_message(self):
        """Sends a message that a new process has been created.
        """
        cmdText = '<process/>'
        cmd = NetCommand(CMD_PROCESS_CREATED, 0, cmdText)
        self.writer.add_command(cmd)

    def send_qt_patch_success_message(self):
        cmdText = str(os.getpid())
        cmd = NetCommand(CMD_QT_PATCH_SUCCESS, 0, cmdText)
        self.writer.add_command(cmd)

    def run(self, file, globals=None, locals=None, is_module=False, set_trace=True):
        module_name = None
        entry_point_fn = ''
        if is_module:
            # When launching with `python -m <module>`, python automatically adds
            # an empty path to the PYTHONPATH which resolves files in the current
            # directory, so, depending how pydevd itself is launched, we may need
            # to manually add such an entry to properly resolve modules in the
            # current directory
            if '' not in sys.path:
                sys.path.insert(0, '')
            file, _, entry_point_fn = file.partition(':')
            module_name = file
            filename = get_fullname(file)
            if filename is None:
                mod_dir = get_package_dir(module_name)
                if mod_dir is None:
                    sys.stderr.write("No module named %s\n" % file)
                    return
                else:
                    filename = get_fullname("%s.__main__" % module_name)
                    if filename is None:
                        sys.stderr.write("No module named %s\n" % file)
                        return
                    else:
                        file = filename
            else:
                file = filename
                mod_dir = os.path.dirname(filename)
                main_py = os.path.join(mod_dir, '__main__.py')
                main_pyc = os.path.join(mod_dir, '__main__.pyc')
                if filename.endswith('__init__.pyc'):
                    if os.path.exists(main_pyc):
                        filename = main_pyc
                    elif os.path.exists(main_py):
                        filename = main_py
                elif filename.endswith('__init__.py'):
                    if os.path.exists(main_pyc) and not os.path.exists(main_py):
                        filename = main_pyc
                    elif os.path.exists(main_py):
                        filename = main_py

            sys.argv[0] = filename

        if os.path.isdir(file):
            new_target = os.path.join(file, '__main__.py')
            if os.path.isfile(new_target):
                file = new_target

        m = None
        if globals is None:
            m = save_main_module(file, 'PyQtInspect.pqi')
            globals = m.__dict__
            try:
                globals['__builtins__'] = __builtins__
            except NameError:
                pass  # Not there on Jython...

        if locals is None:
            locals = globals

        # Predefined (writable) attributes: __name__ is the module's name;
        # __doc__ is the module's documentation string, or None if unavailable;
        # __file__ is the pathname of the file from which the module was loaded,
        # if it was loaded from a file. The __file__ attribute is not present for
        # C modules that are statically linked into the interpreter; for extension modules
        # loaded dynamically from a shared library, it is the pathname of the shared library file.

        # I think this is an ugly hack, bug it works (seems to) for the bug that says that sys.path should be the same in
        # debug and run.
        if sys.path[0] != '' and m is not None and m.__file__.startswith(sys.path[0]):
            # print >> sys.stderr, 'Deleting: ', sys.path[0]
            del sys.path[0]

        if not is_module:
            # now, the local directory has to be added to the pythonpath
            # sys.path.insert(0, os.getcwd())
            # Changed: it's not the local directory, but the directory of the file launched
            # The file being run must be in the pythonpath (even if it was not before)
            sys.path.insert(0, os.path.split(os.path.realpath(file))[0])

        if set_trace:

            while not self.ready_to_run:
                time.sleep(0.1)  # busy wait until we receive run command

        t = threadingCurrentThread()
        thread_id = get_current_thread_id(t)

        if hasattr(sys, 'exc_clear'):
            # we should clean exception information in Python 2, before user's code execution
            sys.exc_clear()

        return self._exec(is_module, entry_point_fn, module_name, file, globals, locals)

    def _exec(self, is_module, entry_point_fn, module_name, file, globals, locals):
        '''
        This function should have frames tracked by unhandled exceptions (the `_exec` name is important).
        '''
        if not is_module:
            execfile(file, globals, locals)  # execute the script
        else:
            # treat ':' as a separator between module and entry point function
            # if there is no entry point we run we same as with -m switch. Otherwise we perform
            # an import and execute the entry point
            if entry_point_fn:
                mod = __import__(module_name, level=0, fromlist=[entry_point_fn], globals=globals, locals=locals)
                func = getattr(mod, entry_point_fn)
                func()
            else:
                # Run with the -m switch
                import runpy
                if hasattr(runpy, '_run_module_as_main'):
                    # Newer versions of Python actually use this when the -m switch is used.
                    if sys.version_info[:2] <= (2, 6):
                        runpy._run_module_as_main(module_name, set_argv0=False)
                    else:
                        runpy._run_module_as_main(module_name, alter_argv=False)
                else:
                    runpy.run_module(module_name)
        return globals

    def exiting(self):
        # noinspection PyBroadException
        try:
            sys.stdout.flush()
        except:
            pass
        # noinspection PyBroadException
        try:
            sys.stderr.flush()
        except:
            pass
        self.check_output_redirect()
        cmd = self.cmd_factory.make_exit_message()
        self.writer.add_command(cmd)

    def send_widget_message(self, widget_info: QWidgetInfo):
        cmd = self.cmd_factory.make_widget_info_message(widget_info)
        self.writer.add_command(cmd)

    # trace_dispatch = _trace_dispatch
    # frame_eval_func = frame_eval_func
    # dummy_trace_dispatch = dummy_trace_dispatch

    # noinspection SpellCheckingInspection
    @staticmethod
    def stoptrace():
        """A proxy method for calling :func:`stoptrace` from the modules where direct import
        is impossible because, for example, a circular dependency."""
        pass

    def dispose_and_kill_all_pqi_threads(self):
        # TODO
        py_db = get_global_debugger()
        if py_db is self:
            set_global_debugger(None)

    def enable_inspect(self, extra_data: OptionalDict = None):
        if extra_data is None:
            extra_data = {}

        self.inspect_enabled = True
        self._inspect_extra_data = extra_data

    def disable_inspect(self):
        self.inspect_enabled = False
        self._inspect_extra_data = {}

    @property
    def mock_left_button_down(self) -> bool:
        return self._inspect_extra_data.get('mock_left_button_down', False)

    def notify_inspect_finished(self, widget):
        self.select_widget(widget)

        cmd = self.cmd_factory.make_inspect_finished_message()
        self.writer.add_command(cmd)

    def select_widget(self, widget):
        self._selected_widget = widget

    def exec_code_in_selected_widget(self, code):
        exec_code_in_widget(self._selected_widget, code)

    def notify_exec_code_result(self, result):
        cmd = self.cmd_factory.make_exec_code_result_message(result)
        self.writer.add_command(cmd)

    def notify_exec_code_error_message(self, err_msg):
        cmd = self.cmd_factory.make_exec_code_err_message(err_msg)
        self.writer.add_command(cmd)

    def notify_thread_not_alive(self, thread_id):
        ...

    def register_widget(self, widget):
        self._id_to_widget[id(widget)] = widget

    def set_widget_highlight_by_id(self, widget_id: int, is_highlight: bool):
        widget = self._id_to_widget.get(widget_id, None)

        if widget is None:
            return

        if not is_wrapped_pointer_valid(widget):
            del self._id_to_widget[widget_id]
            return

        set_widget_highlight(widget, is_highlight)

    def select_widget_by_id(self, widget_id):
        widget = self._id_to_widget.get(widget_id, None)

        if widget is None:
            return

        if not is_wrapped_pointer_valid(widget):
            del self._id_to_widget[widget_id]
            return

        self.select_widget(widget)

    def send_widget_info_to_server(self, widget, extra=None):
        if extra is None:
            extra = {}

        parent_info = list(get_parent_info(widget))
        parent_classes, parent_ids, parent_obj_names = [], [], []
        if parent_info:
            parent_classes, parent_ids, parent_obj_names = zip(*get_parent_info(widget))

        widget_info = QWidgetInfo(
            class_name=widget.__class__.__name__,
            object_name=get_widget_object_name(widget),
            id=id(widget),
            stacks_when_create=filter_trace_stack(getattr(widget, '_pqi_stacks_when_create', [])),
            size=get_widget_size(widget),
            pos=get_widget_pos(widget),
            parent_classes=parent_classes,
            parent_ids=parent_ids,
            parent_object_names=parent_obj_names,
            stylesheet=get_stylesheet(widget),
            extra=extra,
        )
        self.send_widget_message(widget_info)

    def notify_widget_info(self, widget_id, extra):
        widget = self._id_to_widget.get(widget_id, None)

        if widget is None:
            return

        if not is_wrapped_pointer_valid(widget):
            del self._id_to_widget[widget_id]
            return

        self.send_widget_info_to_server(widget, extra)

    def notify_children_info(self, widget_id):
        """
        Notify the children information of the given widget to the PyDB debugger.

        @param widget_id: The ID of the widget.

        @note: used for the bottom hierarchy view of the server GUI program.
        """
        widget = self._id_to_widget.get(widget_id, None)

        if widget is None:
            return

        if not is_wrapped_pointer_valid(widget):
            del self._id_to_widget[widget_id]
            return

        children_info_list = list(get_children_info(widget))
        child_classes, child_ids, child_object_names = [], [], []
        if children_info_list:
            child_classes, child_ids, child_object_names = zip(*children_info_list)

        children_info = QWidgetChildrenInfo(
            widget_id=widget_id,
            child_classes=child_classes,
            child_ids=child_ids,
            child_object_names=child_object_names,
        )

        cmd = self.cmd_factory.make_children_info_message(children_info)
        self.writer.add_command(cmd)


def set_debug(setup):
    import logging

    setup['DEBUG_RECORD_SOCKET_READS'] = True
    setup['LOG_TO_FILE_LEVEL'] = logging.DEBUG
    setup['LOG_TO_CONSOLE_LEVEL'] = logging.DEBUG


# =======================================================================================================================
# settrace
# =======================================================================================================================
def settrace(
        host='127.0.0.1',
        port=19394,
        patch_multiprocessing=False,
        qt_support='pyqt5',
        is_attach=False,
):
    '''Sets the tracing function with the pydev debug function and initializes needed facilities.

    @param host: the user may specify another host, if the debug server is not in the same machine (default is the local
        host)

    @param port: specifies which port to use for communicating with the server (note that the server must be started
        in the same port). @note: currently it's hard-coded at 5678 in the client

    @param patch_multiprocessing: if True we'll patch the functions which create new processes so that launched
        processes are debugged.

    @param qt_support: the Qt support to be used (currently 'pyqt5' is default).

    @param is_attach: if True, we're attaching to an existing process (and not launching a new one) now.
    '''
    _set_trace_lock.acquire()
    try:
        _locked_settrace(
            host,
            port,
            patch_multiprocessing,
            qt_support,
            is_attach
        )
    finally:
        _set_trace_lock.release()


_set_trace_lock = thread.allocate_lock()


def _locked_settrace(
        host,
        port,
        patch_multiprocessing,
        qt_support,
        is_attach
):
    if SetupHolder.setup is None:
        setup = {
            'client': host,  # dispatch expects client to be set to the host address when server is False
            'server': False,
            'port': int(port),
            'multiprocess': patch_multiprocessing,
            'qt-support': qt_support,
            'stack-max-depth': 0,
            'show-pqi-stack': False,
        }
        SetupHolder.setup = setup

    if patch_multiprocessing:
        try:
            import PyQtInspect._pqi_bundle.pqi_monkey
        except:
            pass
        else:
            PyQtInspect._pqi_bundle.pqi_monkey.patch_new_process_functions()

    global connected
    connected = False

    # Reset created PyDB daemon threads after fork - parent threads don't exist in a child process.
    PyDBDaemonThread.created_pydb_daemon_threads = {}

    if not connected:
        debugger = PyDB()
        debugger.connect(host, port)  # Note: connect can raise error.

        # Mark connected only if it actually succeeded.
        connected = True

        while not debugger.ready_to_run:
            time.sleep(0.1)  # busy wait until we receive run command

        # Stop the tracing as the last thing before the actual shutdown for a clean exit.
        # atexit.register(stoptrace)  todo

    try:
        import PyQtInspect._pqi_bundle.pqi_monkey_qt
    except:
        pass
    else:
        PyQtInspect._pqi_bundle.pqi_monkey_qt.patch_qt(qt_support, is_attach=is_attach)


# =======================================================================================================================
# main
# =======================================================================================================================
def usage(do_exit=True, exit_code=0):
    sys.stdout.write('Usage:\n')
    sys.stdout.write(
        '\tpqi.py --port N --client hostname [--multiprocess] [--show-pqi-stack] '
        '--qt-support=[pyqt5|pyqt6|pyside2|pyside6] '
        '--file executable [file_options]\n'
    )
    if do_exit:
        sys.exit(exit_code)


def main():
    # parse the command line. --file is our last argument that is required
    try:
        from PyQtInspect._pqi_bundle.pqi_command_line_handling import process_command_line
        setup = process_command_line(sys.argv)
        SetupHolder.setup = setup
    except ValueError:
        traceback.print_exc()
        return usage(exit_code=1)

    # for debug
    if SHOW_DEBUG_INFO_ENV:
        set_debug(setup)

    DebugInfoHolder.DEBUG_RECORD_SOCKET_READS = setup.get('DEBUG_RECORD_SOCKET_READS',
                                                          DebugInfoHolder.DEBUG_RECORD_SOCKET_READS)
    DebugInfoHolder.LOG_TO_FILE_LEVEL = setup.get('LOG_TO_FILE_LEVEL', DebugInfoHolder.LOG_TO_FILE_LEVEL)
    DebugInfoHolder.LOG_TO_CONSOLE_LEVEL = setup.get('LOG_TO_CONSOLE_LEVEL', DebugInfoHolder.LOG_TO_CONSOLE_LEVEL)

    # connect
    port = setup['port']
    host = setup['client']

    debugger = PyDB()

    try:
        debugger.connect(host, port)
    except:
        sys.stderr.write("Could not connect to %s: %s\n" % (host, port))
        traceback.print_exc()
        sys.exit(1)

    import PyQtInspect._pqi_bundle.pqi_monkey

    if setup['multiprocess']:
        PyQtInspect._pqi_bundle.pqi_monkey.patch_new_process_functions()

    is_module = setup['module']

    enable_qt_support(setup['qt-support'])
    if setup['file']:
        debugger.run(setup["file"], None, None, is_module)


if __name__ == '__main__':
    main()



from setuptools import find_packages, setup


def get_long_description() -> str:
    with open("README.md", encoding='utf8') as file:
        return file.read()

setup(
    name="PyQtInspect",
    version="0.3.4",
    url="https://jeza-chen.com/PyqtInspect",
    author="Jianzhang Chen",
    author_email="jezachen@163.com",
    license="GPLv3",
    description="To inspect PyQt/PySide program elements like Chrome's element inspector",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=("*examples", "*examples.*")),
    python_requires=">=3.7, <4",
    keywords="pyqt pyside inspect",
    install_requires=[
        "psutil",
        "PyQt5",
        "wingrab; platform_system=='Windows'",  # Only available on Windows
    ],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
    entry_points={
        "console_scripts": [
            "pqi-server = PyQtInspect.pqi_server_gui:main",
        ],
    },
    # Use MANIFEST.in to include all files in the pqi_attach directory
    include_package_data=True,
)



# -*- encoding:utf-8 -*-
# ==============================================
# Author: Jeza Chen
# Time: 2023/8/18 14:41
# Description: 
# ==============================================
import sys

IS_PY2 = sys.version_info < (3,)

import threading

import time

import socket

import select

if IS_PY2:
    import thread
    import Queue as _queue
    import xmlrpclib
    import SimpleXMLRPCServer as _pydev_SimpleXMLRPCServer
    import BaseHTTPServer
else:
    import _thread as thread
    import queue as _queue
    try:
        import xmlrpc.client as xmlrpclib
        import xmlrpc.server as _pydev_SimpleXMLRPCServer
    except ImportError:  # maybe some python version does not have xmlrpc.client
        pass
    import http.server as BaseHTTPServer
